@@ -14,10 +14,13 @@ Author:  anatoly techtonik <techtonik@gmail.com>
 License: Public Domain (use MIT if Public Domain doesn't work for you)
 """
 
-__version__ = '1.2'
+__version__ = '1.4'
 
 import os,sys
 
+WINDOWS = os.name == 'nt'
+PY3K = sys.version_info >= (3,)
+
 # Windows constants
 # http://msdn.microsoft.com/en-us/library/ms683231%28v=VS.85%29.aspx
 
@@ -25,7 +28,8 @@ STD_INPUT_HANDLE  = -10
 STD_OUTPUT_HANDLE = -11
 STD_ERROR_HANDLE  = -12
 
-if os.name == 'nt':
+
+if WINDOWS:
     # get console handle
     from ctypes import windll, Structure, byref
     try:
@@ -102,7 +106,7 @@ def getwidth():
     *nix part uses termios ioctl TIOCGWINSZ call.
     """
     width = None
-    if os.name == 'nt':
+    if WINDOWS:
         return _windows_get_window_size()[0]
     elif os.name == 'posix':
         return _posix_get_window_size()[0]
@@ -121,7 +125,7 @@ def getheight():
     *nix part uses termios ioctl TIOCGWINSZ call.
     """
     height = None
-    if os.name == 'nt':
+    if WINDOWS:
         return _windows_get_window_size()[1]
     elif os.name == 'posix':
         return _posix_get_window_size()[1]
@@ -134,12 +138,19 @@ def getheight():
 
 # --- getch() constants and input logic ---
 
-ENTER = '\n'
-ESC = '\x1B'
-LEFT = ['\x1b', '[', 'D']
-RIGHT = ['\x1b', '[', 'C']
-UP = ['\x1b', '[', 'A']
-DOWN = ['\x1b', '[', 'B']
+if WINDOWS:
+    ENTER = '\x0d'
+    LEFT =  ['\xe0', 'K']
+    UP =    ['\xe0', 'H']
+    RIGHT = ['\xe0', 'M']
+    DOWN =  ['\xe0', 'P']
+else:
+    ENTER = '\n'
+    LEFT =  ['\x1b', '[', 'D']
+    UP =    ['\x1b', '[', 'A']
+    RIGHT = ['\x1b', '[', 'C']
+    DOWN =  ['\x1b', '[', 'B']
+ESC = '\x1b'
 
 def dumpkey(key):
     """
@@ -149,12 +160,11 @@ def dumpkey(key):
     def hex3fy(key):
         """Helper to convert string into hex string (Python 3 compatible)"""
         from binascii import hexlify
-        py3k = sys.version_info > (3,)
         # Python 3 strings are no longer binary, encode them for hexlify()
-        if py3k:
+        if PY3K:
            key = key.encode('utf-8')
         keyhex = hexlify(key).upper()
-        if py3k:
+        if PY3K:
            keyhex = keyhex.decode('utf-8')
         return keyhex
     if type(key) == str:
@@ -162,46 +172,67 @@ def dumpkey(key):
     else:
         return ' '.join( [hex3fy(s) for s in key] )
 
+# [ ] recognize multiple-character sequences such as arrow keys
 def getch():
     """
-    Wait for keypress, return character or a list of characters.
+    Wait for keypress, return character or a list of characters that correspond
+    to the pressed key (or keys waiting in the input buffer).
 
-    Arrows and special keys generate a sequence of characters, so if there are
-    extra symbols in input buffer, this function returns list.
+    Arrows and special keys generate a sequence of characters.
     """
     # Credits: Danny Yoo, Python Cookbook
     ch = None
     morech = []
     try:
-        import msvcrt
-        ch = msvcrt.getch()
-        # [ ] TODO - test this on Windows
-        while msvcrt.kbhit():
-            morech.append(msvcrt.getch())
+        if PY3K:
+            from msvcrt import kbhit, getwch as _getch
+        else:
+            from msvcrt import kbhit, getch as _getch
+        ch = _getch()
+        # [ ] deal with buffered output - return when a key is
+        #     recognized or scan code exceeds max len
+        while kbhit():
+            morech.append(_getch())
     except ImportError:
-        ''' we're not on Windows, so we try the Unix-like approach '''
-        import sys, tty, termios
+        ''' we're not on Windows, try Unix-like approach '''
+
+        # --- current algorithm ---
+        # 1. switch to char-by-char input mode
+        # 2. turn off echo
+        # 3. wait for at least one char to appear
+        # 4. read the rest of the character buffer
+        # 5. [ ] return list of characters
+        import sys, termios
+
         fd = sys.stdin.fileno()
-        # save old terminal settings, because we are changing them
+        # save old terminal settings
         old_settings = termios.tcgetattr(fd)
         try:
-            # set terminal to "cbreak" mode, in which driver returns
-            # one char at a time instead of one line at a time
-            #
-            # tty.setcbreak() is just a helper for tcsetattr() call, see
-            # http://hg.python.org/cpython/file/c6880edaf6f3/Lib/tty.py
-            tty.setcbreak(fd)
+            # change terminal settings - turn off canonical mode and echo
+            # in canonical mode read from stdin returns one line at a time
+            # and we need one char at a time (see DESIGN.rst for more info)
+            newattr = list(old_settings)
+            newattr[3] &= ~termios.ICANON
+            newattr[3] &= ~termios.ECHO
+            newattr[6][termios.VMIN] = 1   # block until one char received
+            newattr[6][termios.VTIME] = 0
+            # TCSANOW below means apply settings immediately
+            termios.tcsetattr(fd, termios.TCSANOW, newattr)
+
+            # [ ] this fails when stdin is redirected, like
+            #       ls -la | pager.py
+            #   [ ] also check on Windows
             ch = sys.stdin.read(1)
 
 
-            # clear input buffer placing all available chars into morech
-            newattr = termios.tcgetattr(fd)   # change terminal settings
-                                              # to allow non-blocking read
+            # move rest of chars (if any) from input buffer to `morech`
+            morech = []
+            # change terminal settings - enable non-blocking read
+            newattr = termios.tcgetattr(fd)
             newattr[6][termios.VMIN] = 0      # CC structure
             newattr[6][termios.VTIME] = 0
             termios.tcsetattr(fd, termios.TCSANOW, newattr)
 
-            morech = []
             while True:
                 ch2 = sys.stdin.read(1)
                 if ch2 != '':
@@ -271,7 +302,7 @@ def page(content, pagecallback=prompt):
                 linelist = ['']
             lines2print = min(len(linelist), linesleft)
             for i in range(lines2print):
-                if os.name == 'nt' and len(line) == width:
+                if WINDOWS and len(line) == width:
                     # avoid extra blank line by skipping linefeed print
                     echo(linelist[i])
                 else:
@@ -299,7 +330,7 @@ def page(content, pagecallback=prompt):
 
 # --- Manual tests when pager executed as a module ---
 
-def manual_test_console():
+def _manual_test_console():
     print("\nconsole size: width %s, height %s" % (getwidth(), getheight()))
     echo("--<enter>--")
     getch()
@@ -407,7 +438,7 @@ def manual_test_console():
     print("")
 
 
-def manual_test_getch():
+def _manual_test_getch():
     echo("\n")
     # special keys
     keys = 'ENTER LEFT UP RIGHT DOWN ESC'.split()
@@ -420,16 +451,42 @@ def manual_test_getch():
       else:
         echo("FAILED: getch() returned %s (hex %s)\n" % (key, dumpkey(key)))
 
+
+
 if __name__ == '__main__':
-    print("Manual tests for pager module.")
-    ch = ''
-    while ch != '0':
-      print("\n1. Test output")
-      print("2. Test input")
-      print("0. Exit")
-      ch = getch()
-      if ch == '1':
-        manual_test_console()
-      elif ch == '2':
-        manual_test_getch()
+    # check if pager.py is running in interactive mode
+    # (without stdin redirection)
+    stdin_fd = sys.stdin.fileno()
+    if os.isatty(stdin_fd):
+        if len(sys.argv) == 1:
+            print("pager v%s" % __version__)
+            print("usage: pager.py <file>")
+            print("       pager.py --test\n")
+            sys.exit(-1)
+
+        #       pager.py --test
+        elif sys.argv[1] == '--test':
+            print("Manual tests for pager module.")
+            ch = ''
+            while ch != '0':
+                print("\n1. Test output")
+                print("2. Test input")
+                print("0. Exit")
+                ch = getch()
+                if ch == '1':
+                    _manual_test_console()
+                elif ch == '2':
+                    _manual_test_getch()
+
+        #       pager.py <file>
+        else:
+            with open(sys.argv[1]) as f:
+                page(f)
+
+    else:
+        # mode with piped stdin needs more tuits
+        pass
+        # print("       pager.py < <file>")
+        # print("       <command> | pager.py")
+        #page(sys.stdin)
 
